#include <owl/celarray.h>
#include <owl/gadgetwi.h>
#include <owl/uihelper.h>
#include <winsys/uimetric.h>

#include "fltbtnga.h"
#include "toolhlpr.h"
#include "xpctlbar.h"

const long RopPSDPxax = 0x00B8074AL;
const long RopDSPDxax = 0x00E20746L;

// TFlatButtonGadget implementation
//
TFlatButtonGadget::TFlatButtonGadget(TResId resId, int id, TType type,
                                     bool enabled, TState state,
                                     bool sharedGlyph)
:
 TButtonGadget(resId, id, type, enabled, state, sharedGlyph), mouseIsIn(false),
 hotColor(193, 210, 238), selectColor(152, 181, 226)
{
}

void
TFlatButtonGadget::Paint(TDC& dc)
{
  // Make sure our style matches our state, then paint our gadget borders
  //
  BorderStyle = (Pressed || State == Down) ? ButtonDn : ButtonUp;
  PaintBorder(dc);

  // Determine which CelArray to paint from, ours or our Window's
  //
  TCelArray& celArray = CelArray ? *CelArray : Window->GetCelArray();

  // Get the inner rect to use as the button face
  //
  TRect faceRect;
  GetInnerRect(faceRect);

  // Calculate the source rect from the celarray
  //
  TRect srcRect(celArray.CelRect(GlyphIndex));

  // The dest point of the glyph is relative to the face rect
  //
  TPoint dstPt(BitmapOrigin - faceRect.TopLeft());

  // Paint the button face
  //
  PaintFace(celArray, srcRect, dc, faceRect, dstPt);
}

void
TFlatButtonGadget::PaintBorder(TDC& dc)
{
  if ((GetEnabled() && mouseIsIn) || GetButtonState()) {
    TRect rect(0, 0, Bounds.Width(), Bounds.Height());

    FillSolidRect(dc, rect, (Pressed ? selectColor : hotColor));

    Draw3dRect(dc, rect, TColor::SysActiveCaption, TColor::SysActiveCaption);
  }
}

void
TFlatButtonGadget::BuildMask(TDC& maskDC, const TPoint& maskDst,
                             const TSize& maskSize, TDC& glyphDC,
                             const TRect& glyphRect)
{
  maskDC.PatBlt(0, 0, maskSize.cx, maskSize.cy, BLACKNESS);

  // We use Sys3dFace as the mask color
  //
  TColor bkColor = glyphDC.SetBkColor(TColor::Sys3dFace);
  maskDC.BitBlt(maskDst.x, maskDst.y, maskSize.cx, maskSize.cy,
                glyphDC, glyphRect.left, glyphRect.top, SRCCOPY);
  glyphDC.SetBkColor(bkColor);
}

void
TFlatButtonGadget::DitherBackground(TDC& dc, TDC& maskDC, const TRect& dstRect)
{
  THatch8x8Brush brush(THatch8x8Brush::Hatch11F1,
                       TColor::Sys3dFace,
                       TColor::Sys3dHilight);
  dc.SelectObject(brush);

  TColor txtColor = dc.SetTextColor(TColor::Black);
  TColor bkColor  = dc.SetBkColor  (TColor::White);

  dc.BitBlt(dstRect, maskDC, TPoint(0, 0), RopDSPDxax);

  dc.SetTextColor(txtColor);
  dc.SetBkColor  (bkColor);

  dc.RestoreBrush();
}

void
TFlatButtonGadget::PaintFace(TBitmap& glyph, TRect& srcRect, TDC& dc,
                             TRect& faceRect, TPoint& dstPoint)
{
  // Select our glyph in a working memory dc
  //
  TMemoryDC memDC(glyph);
  memDC.SetTextColor(TColor::Black);
  memDC.SetBkColor  (TColor::White);

  // Calculate the glyph's destination rect
  //
  TRect dstRect(faceRect.TopLeft() + dstPoint, srcRect.Size());
  dstRect &= faceRect;

  // Build the mask
  //
  TBitmap* mask = new TBitmap(faceRect.Width(), faceRect.Height(), 1, 1, 0);
  TMemoryDC maskDC(*mask);
  BuildMask(maskDC, dstPoint, faceRect.Size(), memDC, srcRect);

  // Draw according the specified state
  //
  if (!GetEnabled()) {
    // Inactivate mask - convert the highlight color to 1's on existing mask
    //
    memDC.SetBkColor(TColor::Sys3dHilight);
    maskDC.BitBlt(srcRect, memDC, TPoint(0, 0), SRCPAINT);

    // Gray the image using the standard button shadow color
    //
    dc.SelectObject(TBrush(TColor::Sys3dShadow));

    TColor bkColor = dc.SetBkColor(TColor::Sys3dHilight);

    dc.BitBlt(dstRect, maskDC, dstPoint, RopPSDPxax);

    dc.SetBkColor(bkColor);

    dc.RestoreBrush();
  }
  else {
    // Do a transparent blit using the mask
    //
    const long DSTCOPY = 0x00AA0029;  // Undocumented ROP

    dc.MaskBlt(dstRect, memDC, srcRect.TopLeft(), *mask, dstPoint,
               MAKEROP4(DSTCOPY, SRCCOPY));
  }

  // Do any filling determined by the state arguments
  //
  if (State == Indeterminate)
    // Dither the background everywhere except where the glyph is
    //
    DitherBackground(dc, maskDC, faceRect);

  maskDC.RestoreBitmap();

  delete mask;
}

void
TFlatButtonGadget::MouseEnter(uint modKeys, TPoint& point)
{
  TButtonGadget::MouseEnter(modKeys, point);

  mouseIsIn = true;

  Invalidate();
}

void
TFlatButtonGadget::MouseLeave(uint modKeys, TPoint& point)
{
  mouseIsIn = false;

  TGadget::Invalidate(true);

  TButtonGadget::MouseLeave(modKeys, point);
}

// TFlatSeparatorGadget implementation
//
TFlatSeparatorGadget::TFlatSeparatorGadget(int w, int h, int id)
:
 TGadget(id)
{
  ShrinkWrapWidth = ShrinkWrapHeight = false;

  SetEnabled(false);

  width  = (w ? w : TUIMetric::CxSizeFrame * 2);
  height = (h ? h : TUIMetric::CxSizeFrame * 5);

  Bounds = TRect(0, 0, width, height);
}

void
TFlatSeparatorGadget::Inserted()
{
  BorderStyle = None;
}

void
TFlatSeparatorGadget::Paint(TDC& dc)
{
  TGadgetWindow::TTileDirection direction = Window->GetDirection();

  if (direction == TGadgetWindow::Horizontal) {
    if (Bounds.Width() != width)
      UpdateBounds(width, height);

    PaintVertically(dc);
  }
  else if (direction == TGadgetWindow::Vertical) {
    if (Bounds.Width() != height)
      UpdateBounds(height, width);

    PaintHorizontally(dc);
  }
  else if (Bounds.Height() != 0 || Bounds.Width() != 0)
    UpdateBounds(0, 0);
}

void
TFlatSeparatorGadget::PaintHorizontally(TDC& dc)
{
  TRect rect(0, 0,  Bounds.Width(), Bounds.Height() / 2 + 1);
  TUIBorder(rect, TUIBorder::EdgeEtched, TUIBorder::Bottom).Paint(dc);
}

void
TFlatSeparatorGadget::PaintVertically(TDC& dc)
{
  TRect rect(0, 0,  Bounds.Width() / 2 + 1, Bounds.Height());
  TUIBorder(rect, TUIBorder::EdgeEtched, TUIBorder::Right).Paint(dc);
}

void
TFlatSeparatorGadget::UpdateBounds(int w, int h)
{
  Bounds = TRect(0, 0, w, h);
  Window->LayoutSession();
  Window->Invalidate();
}

// TFlatGripperGadget implementation
//
TFlatGripperGadget::TFlatGripperGadget(int w, int h, int id)
:
 TFlatSeparatorGadget(w, h, id)
{
  SetEnabled(true);
}

void
TFlatGripperGadget::MouseEnter(uint modKeys, TPoint& point)
{
  TFlatSeparatorGadget::MouseEnter(modKeys, point);

  Window->SetCursor(0, IDC_SIZEALL);
}

void
TFlatGripperGadget::MouseLeave(uint modKeys, TPoint& point)
{
  Window->SetCursor(0, IDC_ARROW);

  TFlatSeparatorGadget::MouseLeave(modKeys, point);
}

void
TFlatGripperGadget::LButtonDown(uint modKeys, TPoint& point)
{
  Window->HandleMessage(WM_LBUTTONDOWN, modKeys,
                        MkParam2(int16(point.x), int16(point.y)));
}

void
TFlatGripperGadget::PaintHorizontally(TDC& dc)
{
  TColor gripperColor = DarkenColor(TColor::Sys3dFace, 100);

  int w = Bounds.Width ();
  for (int x = 0; x < w; x += 4)
    FillSolidRect(dc, x, 1, 2, 2, gripperColor);
}

void
TFlatGripperGadget::PaintVertically(TDC& dc)
{
  TColor gripperColor = DarkenColor(TColor::Sys3dFace, 100);

  int h = Bounds.Height();
  for (int y = 0; y < h; y += 4)
    FillSolidRect(dc, 1, y, 2, 2, gripperColor);
}

